/* connection.c */

#include "config.h"

#include "utilities.h"
#include "my_types.h"
#include "network.h"
#include "connection.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <err.h>
#include <errno.h>
#include <sysexits.h>

time_t Curtime;

fd_set Readset;
fd_set Writeset;

int Timeout = 10;
unsigned int Transfer_maxspeed = 100;
unsigned int Overall_send_speed = 0;
unsigned int Overall_recv_speed = 0;
unsigned long Bytes_send[ SENT_TIME ] = { 0, 0, 0, 0 };
unsigned long Bytes_recv[ SENT_TIME ] = { 0, 0, 0, 0 };
const char *Vhost = NULL;

static char sendbuff[BUFFERSIZE];
static struct timeval timestruct;

my_connection_t *
connection_initvalues(void)
{
  my_connection_t *con;

  con = my_calloc(1U, sizeof(my_connection_t));
  con->family = AF_INET;
  con->state = TRANSFER_STATE_UNUSED;
  con->listensocket = FD_UNUSED;
  con->clientsocket = FD_UNUSED;
  con->lastcontact = Curtime;
  return con;
}

void connection_close(my_connection_t * const con, const char * const msg)
{
  if (msg != NULL)
    warnx("Connection closed, %s", msg);

  if (con->state == TRANSFER_STATE_DONE)
    return;

  if (con->listensocket != FD_UNUSED) {
    FD_CLR(con->listensocket, &Readset);
    close(con->listensocket);
    con->listensocket = FD_UNUSED;
  }
  if (con->clientsocket != FD_UNUSED) {
    FD_CLR(con->clientsocket, &Writeset);
    FD_CLR(con->clientsocket, &Readset);
    shutdown_close(con->clientsocket);
    con->clientsocket = FD_UNUSED;
  }
  if (con->file_fd != FD_UNUSED) {
   close(con->file_fd);
   con->file_fd = FD_UNUSED;
  }
  if (con->state == TRANSFER_STATE_LISTENING)
    set_listen_port(con->localport, 0);
  
  con->state = TRANSFER_STATE_DONE;
}

int
connection_open_listen(my_connection_t * const con)
{
  int rc;

  rc = open_listen(con->family, &(con->local), &(con->listensocket), 0, Tcprangestart, 1, NULL);
  if (rc != 0)
    return rc;

  con->connecttime = Curtime;
  con->lastcontact = Curtime;
  con->localport = get_port(&(con->local));
  con->state = TRANSFER_STATE_LISTENING;
  if (Debug > 0)
    warnx("TRANSFER_STATE_LISTENING");
  return 0;
}

static void
connection_setup_complete(my_connection_t * const con)
{
  char *msg;
  SIGNEDSOCK int addrlen;
#if 0
  SIGNEDSOCK int tempi;
#endif
  int tempc;

  if (con->mode == TRANSFER_MODE_SEND) {
#if 0
    tempi = sizeof(int);
    getsockopt(con->clientsocket, SOL_SOCKET, SO_SNDBUF, &tempc, &tempi);
#endif
    tempc = 65535;
    setsockopt(con->clientsocket, SOL_SOCKET, SO_SNDBUF, &tempc, sizeof(int));

#if 0
    getsockopt(con->clientsocket, SOL_SOCKET, SO_SNDLOWAT, &tempc, &tempi);
#endif
    tempc = 24577;
    setsockopt(con->clientsocket, SOL_SOCKET, SO_SNDLOWAT, &tempc, sizeof(int));
  }

  /* Set TOS socket option to max throughput */
  tempc = 0x8; /* IPTOS_THROUGHPUT */
  setsockopt(con->clientsocket, IPPROTO_IP, IP_TOS, &tempc, sizeof(int));

  set_socket_nonblocking(con->clientsocket, 1);

  if (gettimeofday(&timestruct, NULL) < 0)
    err(EX_UNAVAILABLE, "gettimeofday() failed!");
  con->connecttimems = (((unsigned long long)timestruct.tv_sec) * 1000) + (((unsigned long long)timestruct.tv_usec) / 1000);
  con->lastcontact = timestruct.tv_sec;
  con->connecttime = timestruct.tv_sec;

  addrlen = (con->family == AF_INET) ? sizeof(struct sockaddr_in) : sizeof(struct sockaddr_in6);
  
  if ((getpeername(con->clientsocket, &(con->remote.sa), &(addrlen))) < 0)
    warn("Couldn't get Remote IP");
  
  if ((getsockname(con->clientsocket, &(con->local.sa), &(addrlen))) < 0)
    warn("Couldn't get Local IP");
  
   msg = my_calloc(1U, MAXTEXTLENGTH);
   my_getnameinfo(msg, MAXTEXTLENGTH - 1, &(con->remote.sa), 0);
   my_free(&(con->remoteaddr));
   con->remoteaddr = my_strdup(msg);
   my_getnameinfo(msg, MAXTEXTLENGTH - 1, &(con->local.sa), 0);
   my_free(&(con->localaddr));
   con->localaddr = my_strdup(msg);
   my_free(&msg);
}

static void
connection_setup_send(my_connection_t * const con)
{
  struct stat st;

  con->state = TRANSFER_STATE_RUNNING;
  if (Debug > 0)
    warnx("TRANSFER_STATE_RUNNING connection_setup_send()");
  con->file_fd = open(con->file, O_RDONLY);
  if (con->file_fd < 0) {
    warn("Cant Access File '%s' for Send", con->file);
    con->file_fd = FD_UNUSED;
    connection_close(con, "File Error");
    return;
  }

  if (fstat(con->file_fd, &st) < 0) {
    warn("Cant Access File '%s' Details", con->file);
    connection_close(con, "File Error");
    return;
  }
  if (!S_ISREG(st.st_mode)) {
    warnx("Is not a File '%s'", con->file);
    connection_close(con, "File Error");
    return;
  }

  con->file_size = st.st_size;
  if (con->file_size == 0) {
    warn("File '%s' has size of 0 bytes!", con->file);
    connection_close(con, "File Error");
    return;
  }

  connection_setup_complete(con);
  warnx("Connection established (%s -> %s)", con->remoteaddr, con->localaddr);
}
  
void
connection_open_connect(my_connection_t * const con)
{
  SIGNEDSOCK int addrlen;
  int retval;

  bzero ((char *) &(con->remote), sizeof(con->remote));
  con->clientsocket = socket(con->family, SOCK_STREAM, 0);
  if (con->clientsocket < 0) {
    warn("Socket Error");
    return;
  }

  if (con->family == AF_INET) {
    addrlen = sizeof(struct sockaddr_in);
    con->remote.sin.sin_family = AF_INET;
    con->remote.sin.sin_port = htons(con->remoteport);
    retval = inet_pton(AF_INET, con->remoteaddr, &(con->remote.sin.sin_addr));
    if (retval < 0)
      warn("Invalid IP: %s", con->remoteaddr);
  } else {
    addrlen = sizeof(struct sockaddr_in6);
    con->remote.sin6.sin6_family = AF_INET6;
    con->remote.sin6.sin6_port = htons(con->remoteport);
    retval = inet_pton(AF_INET6, con->remoteaddr, &(con->remote.sin6.sin6_addr));
    if (retval < 0)
      warn("Invalid IP: %s", con->remoteaddr);
  }

  bzero((char *) &(con->local), sizeof(con->local));
  con->localport = 0;

  if (bind_local_vhost(&(con->local), con->family, con->clientsocket, Vhost) != 0) {
    connection_close(con, "Couldn't Bind Virtual Host");
    return;
  }

  set_socket_nonblocking(con->clientsocket, 1);

  con->state = TRANSFER_STATE_CONNECTING;
  if (Debug > 0)
    warnx("TRANSFER_STATE_CONNECTING");

  alarm(Timeout);
  addrlen = (con->family == AF_INET) ? sizeof(struct sockaddr_in) : sizeof(struct sockaddr_in6);
  retval = connect(con->clientsocket, &(con->remote.sa), addrlen);
  if ( (retval < 0) && !((errno == EINPROGRESS) || (errno == EAGAIN)) ) {
    warn("Couldn't Connect");
    connection_close(con, "Couldn't Connect");
    alarm(0);
    return;
  }
  alarm(0);

  connection_setup_complete(con);
  warnx("Connection established (%s -> %s)", con->localaddr, con->remoteaddr);

  connection_setup_send(con);
}

static void
connection_setup_recv(my_connection_t * const con)
{
  con->state = TRANSFER_STATE_RUNNING;
  if (Debug > 0)
    warnx("TRANSFER_STATE_RUNNING connection_setup_recv()");
  con->file_fd = open(con->file, O_WRONLY | O_CREAT | O_EXCL, 0640);
  if (con->file_fd < 0) {
    warn("Cant Access File '%s' for Recv", con->file);
    con->file_fd = FD_UNUSED;
    connection_close(con, "File Error");
    return;
  }

  connection_setup_complete(con);
  warnx("Connection established (%s -> %s)", con->remoteaddr, con->localaddr);
}
  
void
connection_accept(my_connection_t * const con)
{
  SIGNEDSOCK int addrlen;

  if (Debug > 0)
    warnx("connection_accept()");
  addrlen = (con->family == AF_INET) ? sizeof(struct sockaddr_in) : sizeof(struct sockaddr_in6);

  con->clientsocket = accept(con->listensocket, &con->local.sa, &addrlen);
  if (con->clientsocket < 0) {
    warn("Accept Error, Aborting Connection");
    connection_close(con, "Connection Error, Try Again");
    return;
  }

  set_listen_port(con->localport, 0);

  FD_CLR(con->listensocket, &Readset);
  close(con->listensocket);
  con->listensocket = FD_UNUSED;

  if (con->mode == TRANSFER_MODE_SEND) {
    connection_setup_send(con);
    return;
  }
  connection_setup_recv(con);
}

void
connection_connected(my_connection_t * const con)
{
  int callval_i;
  int connect_error;
  unsigned int connect_error_len = sizeof(connect_error);

  if (Debug > 0)
    warnx("connection_connected()");
  FD_CLR(con->listensocket, &Readset);
  FD_CLR(con->listensocket, &Writeset);
  callval_i = getsockopt(con->clientsocket,
                         SOL_SOCKET, SO_ERROR,
                         &connect_error, &connect_error_len);

  if (callval_i < 0) {
    warn("Couldn't determine connection status");
    connection_close(con, "Connection failed");
    return;
  }
  if (connect_error) {
    warnc(connect_error, "Connection Failed");
    connection_close(con, "Connection failed");
    return;
  }

  warnx("Connection Established");

  if (con->mode == TRANSFER_MODE_SEND) {
    connection_setup_send(con);
    return;
  }
  connection_setup_recv(con);
}

void
connection_send(my_connection_t * const con)
{
  ssize_t howmuch, howmuch2;
  size_t attempt;
  char *dataptr;
  off_t offset;
  off_t sum;
  int overlimit;

  if (con->mode != TRANSFER_MODE_SEND)
    return;

  if (Transfer_maxspeed > 0) {
    if (con->tx_bucket < TXSIZE) {
      con->overlimit = 1;
      return; /* over transfer limit */
    }
  } else {
    con->tx_bucket = TXSIZE * MAXTXPERLOOP;
  }
 
  sum = Bytes_send[Curtime % SENT_TIME]
      + Bytes_send[(Curtime-1) % SENT_TIME]
      + Bytes_send[(Curtime-2) % SENT_TIME]
      + Bytes_send[(Curtime-3) % SENT_TIME];
  overlimit = Overall_send_speed && (sum >= (Overall_send_speed * 256)); /* 1024 div 4 */

  if (overlimit)
    return; /* over overall limit */
 
  con->overlimit = 0;
  attempt = con->tx_bucket - (con->tx_bucket % TXSIZE);
  if (attempt > BUFFERSIZE)
    attempt = BUFFERSIZE;

  if (con->file_fd_location != con->bytessent) {
    offset = lseek(con->file_fd, con->bytessent, SEEK_SET);
    if (offset != con->bytessent) {
      warn("Can't seek location in file '%s'", con->file);
      connection_close(con, "Unable to locate data in file");
      return;
    }
    con->file_fd_location = con->bytessent;
  }

  do {
    dataptr = sendbuff;
    howmuch = read(con->file_fd, dataptr, attempt);
    if ((howmuch < 0) && (errno != EAGAIN)) {
      warn("Can't read data from file '%s'", con->file);
      connection_close(con, "Unable to read data from file");
      return;
    }
    if (howmuch <= 0)
      break;

    con->file_fd_location += howmuch;
    howmuch2 = write(con->clientsocket, dataptr, howmuch);
    if ((howmuch2 < 0) && (errno != EAGAIN)) {
      warn("Connection Lost");
      connection_close(con, "Connection Lost");
      return;
    }

    if (howmuch2 < 0)
      howmuch2 = 0;

    if (howmuch2 > 0)
      con->lastcontact = Curtime;

    Bytes_send[Curtime % SENT_TIME] += howmuch2;
    con->bytessent += howmuch2;
    con->tx_bucket -= howmuch2;
    sum += howmuch2;

    if (Debug > 0)
      warnx("connection_send()");

    /* if over 50% send one to be fair */
    if (Overall_send_speed && ((sum * 2) >= (Overall_send_speed * 256))) /* 1024 div 4 */
      break;

  } while ((con->tx_bucket >= TXSIZE) && (howmuch2 > 0));

  if (con->bytessent >= con->file_size) {
    con->state = TRANSFER_STATE_WAITING;
    if (Debug > 0)
      warnx("TRANSFER_STATE_RUNNING");
  }
}

static void
connection_complete(my_connection_t * const con)
{
  unsigned long long curtimems;
  unsigned long long timetookms;
  char *tempstr;
  size_t len;

  if (gettimeofday(&timestruct, NULL) < 0)
    err(EX_UNAVAILABLE, "gettimeofday() failed!");
  curtimems = (((unsigned long long)timestruct.tv_sec) * 1000) + (((unsigned long long)timestruct.tv_usec) / 1000);
  timetookms = curtimems - con->connecttimems;
  if (timetookms < 1)
    timetookms = 1;

  tempstr = my_calloc(1U, MAXTEXTLENGTH);
  len = 0;
  if (timetookms > (60*60*1000))
    len += snprintf(tempstr + len, MAXTEXTLENGTH - len - 1,
             " %llu hr", timetookms/60/60/1000);
  if ((timetookms%(60*60*1000)) > (60*1000))
    len += snprintf(tempstr + len, MAXTEXTLENGTH - len - 1,
             " %llu min", (timetookms%(60*60*1000))/60/1000);
  len += snprintf(tempstr + len, MAXTEXTLENGTH - len - 1,
           " %llu.%03llu sec", (timetookms%(60*1000))/1000, (timetookms%1000));

  warnx("Transfer of File '%s' Completed (%llu KB,%s, %0.1f KB/sec)",
        con->file, con->file_size/1024, tempstr,
          ((float)(con->file_size))/1024.0/((float)timetookms/1000.0));
  my_free(&tempstr);

  FD_CLR(con->clientsocket, &Writeset);
  FD_CLR(con->clientsocket, &Readset);
  shutdown_close(con->clientsocket);
  close(con->file_fd);
  con->file_fd = FD_UNUSED;
  con->file_fd_location = 0;
  con->state = TRANSFER_STATE_DONE;
}

void
connection_recv(my_connection_t * const con)
{
  off_t sum;
  ssize_t howmuch, howmuch2;
  int i;
  char buffer[4];

  if (Debug > 0)
    warnx("connection_recv()");

  if (con->mode == TRANSFER_MODE_SEND) {
    howmuch = read(con->clientsocket, sendbuff, BUFFERSIZE);
    if (howmuch < 0) {
      if (errno == EAGAIN)
        return;
      warn("Connection Lost");
    }
    if (howmuch < 1) {
      connection_close(con, "Connection Lost");
      return;
    }
    con->lastcontact = Curtime;
    con->bytesgot += howmuch;

    if (sendbuff[0] == '0') {
      con->lastack = con->file_size;
    }
    return;
  }
  
  howmuch = BUFFERSIZE;
  howmuch2 = 1;
  for (i=0; i<MAXTXPERLOOP; i++) {
    if ((howmuch == BUFFERSIZE && howmuch2 > 0) && is_fd_readable(con->clientsocket)) {
      if (Overall_recv_speed) {
        sum = Bytes_recv[Curtime % SENT_TIME]
            + Bytes_recv[(Curtime-1) % SENT_TIME]
            + Bytes_recv[(Curtime-2) % SENT_TIME]
            + Bytes_recv[(Curtime-3) % SENT_TIME];
        if (sum >= (Overall_recv_speed*4096)) /* 1024 times 4 */
          return;
      }

      /* read more later */
      if ( time(NULL) > Curtime + 3 )
        return;
  
      howmuch  = read(con->clientsocket, sendbuff, BUFFERSIZE);
      if (howmuch < 0) {
        if (errno == EAGAIN)
          return;
        warn("Connection Lost");
      }
      if (howmuch < 1) {
        connection_close(con, "Connection Lost");
        return;
      }

      howmuch2 = write(con->file_fd, sendbuff, howmuch);
      if (howmuch2 != howmuch) {
        warn("Unable to write data to file '%s'", con->file);
        connection_close(con, "Unable to write data to file");
        return;
      }

      con->lastcontact = Curtime;
      con->bytesgot += howmuch2;
      Bytes_recv[Curtime % SENT_TIME] += howmuch2;
    }

    if (con->bytesgot >= con->file_size) {
      buffer[0] = '0';
      if (is_fd_writeable(con->clientsocket))
        write(con->clientsocket, buffer, 1);

      connection_complete(con);
      return;
    }
  }
}

void
connection_flush(my_connection_t * const con)
{
  if (Debug > 0)
    warnx("connection_flush()");
  if (con->lastack < con->file_size)
    return; /* wait */

  connection_complete(con);
}

/* eof */
