/* client.c */

#include "config.h"

#include "utilities.h"
#include "my_types.h"
#include "network.h"
#include "connection.h"

#include <stdlib.h>
#include <unistd.h>
#include <sysexits.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <err.h>
#include <sys/time.h>

my_connection_t *Recv_con = NULL;
my_connection_t *Send_con = NULL;

static void
do_listen(const char *filename)
{
  Recv_con = connection_initvalues();
  Recv_con->mode = TRANSFER_MODE_RECV;
  if (connection_open_listen(Recv_con))
    return;
  /* Success */
  Recv_con->file = strdup(filename);
  Recv_con->file_size = 24414932;
}

static void
send_file(const char *filename)
{
  Send_con = connection_initvalues();
  Send_con->mode = TRANSFER_MODE_SEND;
  Send_con->file = strdup(filename);
  Send_con->remoteport = Tcprangestart;
  Send_con->remoteaddr = strdup("127.0.0.1");

  connection_open_connect(Send_con);
}

static int
set_select_fdset(int highests, my_connection_t *con, int overlimit)
{
  if (con == NULL)
    return highests;

  switch (con->state) {
  case TRANSFER_STATE_LISTENING:
    FD_SET(con->listensocket, &Readset);
    if (con->listensocket > highests)
      highests = con->listensocket;
    break;
  case TRANSFER_STATE_CONNECTING:
    FD_SET(con->clientsocket, &Writeset);
    if (con->clientsocket > highests)
      highests = con->clientsocket;
    break;
  case TRANSFER_STATE_RUNNING:
    if (!overlimit) {
      FD_SET(con->clientsocket, &Writeset);
      if (con->clientsocket > highests)
        highests = con->clientsocket;
    }
    FD_SET(con->clientsocket, &Readset);
    if (con->clientsocket > highests)
      highests = con->clientsocket;
    break;
  case TRANSFER_STATE_WAITING:
    FD_SET(con->clientsocket, &Readset);
    if (con->clientsocket > highests)
      highests = con->clientsocket;
    break;
  default:
    break;
  }
  return highests;
}

static void
do_perform(my_connection_t *con)
{
  if (con == NULL)
    return;

  switch (con->state) {
  case TRANSFER_STATE_RUNNING:
    if (FD_ISSET(con->clientsocket, &Writeset))
      connection_send(con);
    if (FD_ISSET(con->clientsocket, &Readset))
      connection_recv(con);
    break;
  case TRANSFER_STATE_LISTENING:
    if (FD_ISSET(con->listensocket, &Readset))
      connection_accept(con);
    return; /* skip timeout */
  case TRANSFER_STATE_CONNECTING:
    if (con->mode == TRANSFER_MODE_SEND) {
      if (FD_ISSET(con->clientsocket, &Writeset))
        connection_connected(con);
    } else {
      if (FD_ISSET(con->clientsocket, &Readset))
        connection_connected(con);
    }
    break;
  case TRANSFER_STATE_WAITING:
    if (FD_ISSET(con->clientsocket, &Readset))
      connection_recv(con);
    connection_flush(con);
    break;
  case TRANSFER_STATE_DONE:
    my_free( &(con->localaddr) );
    my_free( &(con->remoteaddr) );
    my_free( &(con->file) );
    return; /* skip timeout */
  case TRANSFER_STATE_UNUSED:
    return; /* skip timeout */
  default:
    break;
  }

  if (con->lastcontact + Timeout < Curtime) {
    connection_close(con, "Connection Timed Out");
    return;
  }
}

static void
do_io(void)
{
  static int first_loop = 1;
  static struct timeval timestruct;
  static time_t lasttime;

  unsigned long sum;
  long new_bucket;
  int highests;
  int overlimit;

  if (first_loop) {
    lasttime = Curtime - 1;
  }

  for (;;) {
    FD_ZERO(&Readset);
    FD_ZERO(&Writeset);
    highests = 0;

    sum = Bytes_send[Curtime % SENT_TIME]
        + Bytes_send[(Curtime-1) % SENT_TIME]
        + Bytes_send[(Curtime-2) % SENT_TIME]
        + Bytes_send[(Curtime-3) % SENT_TIME];
    overlimit = Overall_send_speed && (sum >= (Overall_send_speed * 1024));

    highests = set_select_fdset(highests, Recv_con, 0);
    highests = set_select_fdset(highests, Send_con, overlimit);
    timestruct.tv_sec = 0;
    timestruct.tv_usec = 250 * 1000;
    if (select(highests+1, &Readset, &Writeset, NULL, &timestruct) < 0) {
      if (errno != EINTR) {
        warn("Select returned an error");
        usleep(10000); /* prevent fast spinning */
      }
      /* data is undefined on error, zero and continue */
      FD_ZERO(&Readset);
      FD_ZERO(&Writeset);
      continue;
    }
    if (gettimeofday(&timestruct, NULL) < 0)
      err(EX_UNAVAILABLE, "gettimeofday() failed!");
    Curtime = timestruct.tv_sec;
    if (Curtime != lasttime) {
      if (Curtime < lasttime-3) {
        warn("System Time Changed Backwards %lim %lis!!\n",
             (long)(lasttime - Curtime) / 60, (long)(lasttime - Curtime) % 60);
      }
      if (Curtime > lasttime+3) {
        warn("System Time Changed Forward %lim %lis!!\n",
             (long)(lasttime - Curtime) / 60, (long)(lasttime - Curtime) % 60);
      }
      lasttime = Curtime;
      Bytes_send[(Curtime+1) % SENT_TIME] = 0;
      Bytes_recv[(Curtime+1) % SENT_TIME] = 0;
      if (Send_con) {
        if (Transfer_maxspeed) {
          new_bucket = Transfer_maxspeed * 1024;
          Send_con->tx_bucket += new_bucket;
          new_bucket += 1024;
          if (Send_con->tx_bucket > new_bucket)
            Send_con->tx_bucket = new_bucket;
        }
      }
    }
    do_perform(Recv_con);
    do_perform(Send_con);
  }
}

static void
usage(void)
{
  warnx(
"\n"
" Usage: client [options] [...] commands [...]\n"
"\n"
" options:\n"
"-debug            show more info\n"
"-transfermax i    set transfer max in kb/s\n"
"\n"
" commands:\n"
"+send filename    send file\n"
"+listen           recv file\n"
"\n");
  exit(EX_USAGE);
}

#define GET_NEXT_DATA(x)	{ argv++; argc--; x = *argv; \
				if (x == NULL) usage(); }

static void
command_run(int argc, const char *const *argv)
{
  const char *cptr;
  char ch;

  while (--argc > 0) {
    cptr = *(++argv);
    /* options */
    if (*cptr == '-') {
      ch = *(++cptr);
      switch (ch) {
      case 'd':
        Debug ++;
        break;
      case 'r':
        GET_NEXT_DATA(cptr);
        Overall_recv_speed = atoul(cptr);
        break;
      case 't':
        GET_NEXT_DATA(cptr);
        Transfer_maxspeed = atoul(cptr);
        break;
      default:
        usage();
      }
      continue;
    }
    /* commands */
    if (*cptr == '+') {
      ch = *(++cptr);
      switch (ch) {
      case 'l': /* listen */
        GET_NEXT_DATA(cptr);
        do_listen(cptr);
        break;
      case 's': /* send */
        GET_NEXT_DATA(cptr);
        send_file(cptr);
        break;
      default:
        usage();
      }
      continue;
    }
    usage();
  }
}

int
main(int argc, const char *const *argv)
{
  Curtime = time(NULL);
  command_run(argc, argv);
  do_io();
  exit(EX_OK);
}

/* eof */
