/* network.c */

#include "config.h"

#include "utilities.h"
#include "my_types.h"
#include "network.h"

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <err.h>
#include <fcntl.h>
#include <netdb.h>

int Tcprangestart = 4000;
int Tcprangelimit = 65635;
int Max_ports = 100;
int Debug = 0;

static int *port_range = NULL;

int
get_port(my_sockaddr_union_t * const listenaddr)
{
  if (listenaddr->sa.sa_family == AF_INET)
    return ntohs(listenaddr->sin.sin_port);

  return ntohs(listenaddr->sin6.sin6_port);
}

const char *
my_ip_show(char *buffer, size_t len, my_sockaddr_union_t * const sa)
{
  if (sa->sa.sa_family == AF_INET)
    return inet_ntop(sa->sa.sa_family, &(sa->sin.sin_addr.s_addr), buffer, len);
 
  return inet_ntop(sa->sa.sa_family, &(sa->sin6.sin6_addr), buffer, len);
}

int
my_getnameinfo(char *buffer, size_t len, const struct sockaddr * const sa, socklen_t salen)
{
#if !defined(NO_GETADDRINFO)
  char hbuf[NI_MAXHOST], sbuf[NI_MAXSERV];

  if (salen == 0)
    salen = (sa->sa_family == AF_INET) ? sizeof(struct sockaddr_in) : sizeof(struct sockaddr_in6);
  if (getnameinfo(sa, salen, hbuf, sizeof(hbuf), sbuf,
                  sizeof(sbuf), NI_NUMERICHOST | NI_NUMERICSERV)) {
    return snprintf(buffer, len, "(unknown)" );
  }
  return snprintf(buffer, len, "host=%s port=%s", hbuf, sbuf);
#else /* NO_GETADDRINFO */
  const struct sockaddr_in *remoteaddr = (const struct sockaddr_in *)sa;
  unsigned long to_ip = ntohl(remoteaddr->sin_addr.s_addr);
  return snprintf(buffer, len, "%lu.%lu.%lu.%lu:%d",
                  (to_ip >> 24) & 0xFF,
                  (to_ip >> 16) & 0xFF,
                  (to_ip >>  8) & 0xFF,
                  (to_ip      ) & 0xFF,
                  ntohs(remoteaddr->sin_port));
#endif /* NO_GETADDRINFO */
}

int
set_port(my_sockaddr_union_t * const listenaddr, int port)
{
  if (listenaddr->sa.sa_family == AF_INET)
    return ntohs(listenaddr->sin.sin_port = htons(port));

  return ntohs(listenaddr->sin6.sin6_port = htons(port));
}

static int
setup_vhost(my_sockaddr_union_t * const listenaddr, int family, const char * const vhost)
{
  SIGNEDSOCK int addrlen;
  int e;

  if (vhost == NULL)
    return 1;

  if (family == AF_INET) {
    addrlen = sizeof(struct sockaddr_in);
    listenaddr->sin.sin_family = AF_INET;
    e = inet_pton(family, vhost, &(listenaddr->sin.sin_addr));
  } else {
    addrlen = sizeof(struct sockaddr_in6);
    listenaddr->sin6.sin6_family = AF_INET6;
    e = inet_pton(family, vhost, &(listenaddr->sin6.sin6_addr));
  }

  if (e != 1) {
    warnx("Invalid IP: %s", vhost);
    return 1;
  }

  return 0;
}

int
bind_local_vhost(my_sockaddr_union_t * const localaddr, int family, int clientsocket, const char * const vhost)
{
  SIGNEDSOCK int addrlen;
  int e;

  if (!vhost)
    return 0;

  bzero((char*)localaddr, sizeof(my_sockaddr_union_t));
  if (family == AF_INET ) {
    addrlen = sizeof(struct sockaddr_in);
    localaddr->sin.sin_family = AF_INET;
    localaddr->sin.sin_port = htons(0);
    e = inet_pton(family, vhost, &(localaddr->sin.sin_addr));
  } else {
    addrlen = sizeof(struct sockaddr_in6);
    localaddr->sin6.sin6_family = AF_INET6;
    localaddr->sin6.sin6_port = htons(0);
    e = inet_pton(family, vhost, &(localaddr->sin6.sin6_addr));
  }

  if (e != 1) {
    warnx("Invalid IP: %s", vhost);
    return 1;
  }

  if (bind(clientsocket, &(localaddr->sa), addrlen) < 0)
    return 1;

  return 0;
}

void
shutdown_close(int handle)
{
  /*
   * cygwin close() is broke, if outstanding data is present
   * it will block until the TCP connection is dead, sometimes
   * upto 10-20 minutes, calling shutdown() first seems to help
   */
  shutdown(handle, SHUT_RDWR);
  close(handle);
}

int
bind_listen_socket(int fd, my_sockaddr_union_t * const sa)
{
  int retry;
  uint16_t port;
  SIGNEDSOCK int addrlen;

  if (port_range == NULL)
    port_range = my_calloc((unsigned)Max_ports, sizeof(int));

  addrlen = (sa->sa.sa_family == AF_INET) ? sizeof(struct sockaddr_in) : sizeof(struct sockaddr_in6);

  for (retry = 0; retry < Max_ports; retry++) {
    if (Tcprangestart) {
      /* set port number */
      port = Tcprangestart + retry;
      if (port > Tcprangelimit) {
        /* give up */
        retry = Max_ports;
        break;
      }
      if (port_range[ port - Tcprangestart ])
        continue;

      set_port(sa, port);
    } else {
      set_port(sa, 0);
    }

    if (bind(fd, &(sa->sa), addrlen) < 0) {
      if (!Tcprangestart) {
        /* give up */
        retry = Max_ports;
        break;
      }
      continue;
    }
    break;
  }

  if (retry == Max_ports)
    return -1;
 
  if ((getsockname(fd, &(sa->sa), &addrlen)) < 0) {
    warn("Couldn't get Port Number, Aborting");
    return -1;
  }
 
  port = get_port(sa);
  if (Debug > 0)
    warnx("found free port %d", port);
 
  port_range[ port - Tcprangestart ] = 1;
  return 0;
}

void
set_listen_port(int port, int val)
{
  if (port_range == NULL)
    return;

  port_range[ port - Tcprangestart ] = val;
}

int
open_listen(int family, my_sockaddr_union_t * const listenaddr, int *listen_socket, int port, int reuse, int search, const char * const vhost)
{
  int rc;
  int tempc;
  SIGNEDSOCK int addrlen;

  if (vhost != NULL) {
    family = strchr(vhost, ':') ? AF_INET6 : AF_INET;
  }
  *listen_socket = socket(family, SOCK_STREAM, 0);
  if (*listen_socket < 0) {
    warn("Could Not Create Socket, Aborting");
    *listen_socket = FD_UNUSED;
    return 1;
  }

  if (reuse) {
    tempc = 1;
    setsockopt(*listen_socket, SOL_SOCKET, SO_REUSEADDR, &tempc, sizeof(int));
  }

  bzero((char *) listenaddr, sizeof(my_sockaddr_union_t));
  if (family == AF_INET) {
    addrlen = sizeof(struct sockaddr_in);
    listenaddr->sin.sin_family = AF_INET;
    listenaddr->sin.sin_addr.s_addr = INADDR_ANY;
    listenaddr->sin.sin_port = htons(port);
  } else {
    addrlen = sizeof(struct sockaddr_in6);
    listenaddr->sin6.sin6_family = AF_INET6;
    listenaddr->sin6.sin6_port = htons(port);
  }

  setup_vhost(listenaddr, family, vhost);

  if (search) {
    rc = bind_listen_socket(*listen_socket, listenaddr);
  } else {
    rc = bind(*listen_socket, &(listenaddr->sa), addrlen);
  }

  if (rc < 0) {
    warn("Couldn't Bind to Socket, Aborting");
    shutdown_close(*listen_socket);
    *listen_socket = FD_UNUSED;
    return 1;
  }

  if (listen(*listen_socket, 1) < 0) {
    warn("Couldn't Listen, Aborting");
    shutdown_close(*listen_socket);
    *listen_socket = FD_UNUSED;
    return 1;
  }

  return 0;
}

int
set_socket_nonblocking (int s, int nonblock)
{
  long current;

  current = fcntl(s, F_GETFL, 0);
  if (current == -1)
    return -1;

  if (nonblock)
    return fcntl(s, F_SETFL, current | O_NONBLOCK);
  else
    return fcntl(s, F_SETFL, current & ~O_NONBLOCK);
}

/* eof */
